from dataclasses import dataclass
from typing import Any
from typing import Dict


@dataclass()
class TrainingResults:
    """
    Training results data class

    Attributes
    ----------

    model_path: `str`
        The trained model path
    metrics: `Dict[str, Any]`
        Related training metrics

    """

    model_path: str
    metrics: Dict[str, Any]



import os
import sys
from tempfile import mkdtemp
from urllib.parse import urlparse

import mlflow
import pytest
from mlflow.utils import mlflow_tags

from biome.text import AllenNLPTrainerConfiguration
from biome.text import Pipeline
from biome.text import PipelineConfiguration
from biome.text.loggers import MlflowLogger
from biome.text.loggers import WandBLogger
from biome.text.modules.heads import TaskHeadConfiguration
from biome.text.modules.heads import TextClassification
from biome.text.training_results import TrainingResults


@pytest.fixture
def pipeline() -> Pipeline:
    return Pipeline.from_config(
        PipelineConfiguration(
            name="test-pipeline",
            head=TaskHeadConfiguration(type=TextClassification, labels=["A", "B"]),
        )
    )


def test_mlflow_logger(pipeline):

    logger = MlflowLogger(
        experiment_name="test-experiment", run_name="test_run", tag1="my-tag"
    )

    trainer = AllenNLPTrainerConfiguration()

    logger.init_train(pipeline, trainer, training=None)
    for epoch in range(0, 10):
        logger.log_epoch_metrics(epoch, metrics={"key": 10 * epoch})

    model_path = mkdtemp()
    metrics = {"metric": 200}
    logger.end_train(TrainingResults(model_path, metrics))

    run = mlflow.get_run(logger._run_id)
    assert run
    # Tags
    assert "test_run" == run.data.tags[mlflow_tags.MLFLOW_RUN_NAME]
    assert "my-tag" == run.data.tags["tag1"]
    # Parameters
    expected_parmams = {
        "pipeline.features.word.trainable": "True",
        "pipeline.num_parameters": "202",
        "pipeline.num_trainable_parameters": "202",
        "pipeline.features.word.embedding_dim": "50",
        "pipeline.head.type": "biome.text.modules.heads.classification.text_classification.TextClassification",
        "pipeline.head.labels": "['A', 'B']",
        "pipeline.name": "test-pipeline",
        "pipeline.tokenizer.lang": "en",
        "trainer.batch_size": "16",
        "trainer.validation_metric": "-loss",
        "trainer.optimizer.type": "adam",
        "trainer.patience": "2",
        "trainer.num_epochs": "20",
        "trainer.num_serialized_models_to_keep": "1",
        "pipeline.tokenizer.remove_space_tokens": "True",
    }
    assert expected_parmams == run.data.params
    # Artifacts
    assert os.path.basename(model_path) in os.listdir(
        urlparse(run.info.artifact_uri).path
    )
    # Metrics
    for metric in metrics:
        assert (
            metric in run.data.metrics and run.data.metrics[metric] == metrics[metric]
        )


@pytest.mark.skipif("wandb" not in sys.modules, reason="wandb client not installed")
class TestWandBLogger:
    def test_wandb_logger_init_warning(self, monkeypatch, capsys):
        import wandb

        monkeypatch.setattr(wandb.sdk.internal.internal_api.Api, "api_url", None)

        wandb.ensure_configured()
        WandBLogger(run_name="test_run", tags=["test_tag"])
        assert capsys.readouterr().err.startswith("wandb")

    def test_wandb_logger_init_train(
        self, monkeypatch, pipeline, change_to_tmp_working_dir
    ):
        # It's impossible to test wandb ... but these tests will soon go away when we have our lightning trainer ready
        class MockRun:
            def __init__(self, name, tags, dir, project, *args, **kwargs):
                self.name = name
                self.tags = tags
                self.dir = dir
                self.project = project

        def mock_wandb_init(*args, **kwargs):
            return MockRun(*args, **kwargs)

        import wandb

        monkeypatch.setattr(wandb, "init", mock_wandb_init)
        monkeypatch.setenv("WANDB_PROJECT", "mock_project")

        wandb.ensure_configured()
        logger = WandBLogger(run_name="test_run", tags=["test_tag"])

        logger.init_train(
            pipeline=pipeline,
            trainer_configuration=AllenNLPTrainerConfiguration(),
            training=None,
        )

        assert logger._run.name == "test_run"
        assert logger._run.tags[0] == "test_tag"
        assert logger._run.dir.startswith(".wandb")
        assert logger._run.project == "mock_project"



import pytest

from biome.text._helpers import PipelineTrainer
from biome.text.training_results import TrainingResults


@pytest.fixture
def deactivate_pipeline_trainer(monkeypatch):
    def mock_init(*args, **kwargs):
        pass

    monkeypatch.setattr(PipelineTrainer, "__init__", mock_init)

    def mock_train(*args, **kwargs):
        return TrainingResults(
            model_path="mock_output_path", metrics={"mock_metric": 0.0}
        )

    monkeypatch.setattr(PipelineTrainer, "train", mock_train)



from pathlib import Path

import pytest
from numpy.testing import assert_allclose

from biome.text import Dataset
from biome.text import Pipeline
from biome.text.configuration import AllenNLPTrainerConfiguration
from biome.text.configuration import FindLRConfiguration


@pytest.fixture
def train_data_source() -> Dataset:
    """Creates the training dataset"""
    resources_path = Path(__file__).parent.parent / "resources" / "data"
    training_ds = Dataset.from_csv(
        paths=str(resources_path / "business.cat.2k.train.csv")
    )

    return training_ds


@pytest.fixture
def pipeline_dict() -> dict:
    """Pipeline config dict. Updating the labels is needed"""

    pipeline_dict = {
        "name": "german_business_names",
        "features": {
            "word": {
                "embedding_dim": 16,
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
            "pooler": {
                "type": "gru",
                "num_layers": 1,
                "hidden_size": 16,
                "bidirectional": True,
            },
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [16],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_config() -> AllenNLPTrainerConfiguration:
    """Returns the trainer configuration"""
    return AllenNLPTrainerConfiguration(
        batch_size=64, num_epochs=5, optimizer={"type": "adam", "lr": 0.01}
    )


@pytest.fixture
def find_lr_config() -> FindLRConfiguration:
    """Returns FindLRConfiguration function"""
    return FindLRConfiguration(num_batches=11)


def test_find_lr(train_data_source, pipeline_dict, trainer_config, find_lr_config):
    """Asserting that all metrics of the find_lr method match in a controlled enviroment"""

    # Creation of pipeline and vocabulary
    pl = Pipeline.from_config(pipeline_dict)

    # Test prediction
    prev_prediction = pl.predict("test")

    # Find_lr method
    learning_rates, losses = pl.find_lr(
        trainer_config=trainer_config,
        find_lr_config=find_lr_config,
        training_data=train_data_source,
    )

    assert len(learning_rates) == len(losses) == 12
    assert_allclose(
        prev_prediction["probabilities"], pl.predict("test")["probabilities"], rtol=1e-6
    )



import os
import random
import tempfile
from typing import Tuple

import numpy as np
import pytest
import torch
from pytorch_lightning.loggers import CSVLogger
from pytorch_lightning.loggers import LoggerCollection
from pytorch_lightning.loggers import MLFlowLogger
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.loggers import WandbLogger

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import VocabularyConfiguration
from biome.text.configuration import CharFeatures
from biome.text.configuration import TrainerConfiguration
from biome.text.configuration import WordFeatures


@pytest.fixture
def train_valid_dataset(resources_data_path) -> Tuple[Dataset, Dataset]:
    """Returns both training and validation datasets"""

    training_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.train.csv")
    )
    validation_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.valid.csv")
    )

    return training_ds, validation_ds


@pytest.fixture
def pipeline_dict() -> dict:
    return {
        "name": "german_business_names",
        "features": {
            "word": {"embedding_dim": 16, "lowercase_tokens": True},
            "char": {
                "embedding_dim": 16,
                "encoder": {
                    "type": "gru",
                    "num_layers": 1,
                    "hidden_size": 32,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
            "pooler": {
                "type": "gru",
                "num_layers": 1,
                "hidden_size": 16,
                "bidirectional": True,
            },
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [16],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }


def test_text_classification(tmp_path, pipeline_dict, train_valid_dataset):
    """Apart from a well specified training, this also tests the vocab creation!"""

    random.seed(43)
    np.random.seed(43)
    torch.manual_seed(43)

    pl = Pipeline.from_config(pipeline_dict)
    train_ds = train_valid_dataset[0]
    valid_ds = train_valid_dataset[1]

    vocab_config = VocabularyConfiguration(max_vocab_size={"word": 50})
    trainer_config = TrainerConfiguration(
        batch_size=64,
        optimizer={"type": "adam", "lr": 0.01},
        max_epochs=5,
        default_root_dir=str(tmp_path),
        gpus=0,  # turn off gpus even if available
    )

    trainer = Trainer(
        pipeline=pl,
        train_dataset=train_ds,
        valid_dataset=valid_ds,
        trainer_config=trainer_config,
        vocab_config=vocab_config,
    )

    trainer.fit(tmp_path / "output")

    assert pl.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl.vocab.get_vocab_size(CharFeatures.namespace) == 83

    assert pl.num_trainable_parameters == 22070

    evaluation = pl.evaluate(valid_ds)

    assert evaluation["loss"] == pytest.approx(0.8479013895988464, abs=0.003)

    Pipeline.from_pretrained(str(tmp_path / "output" / "model.tar.gz"))


def test_default_root_dir(
    change_to_tmp_working_dir, pipeline_dict, train_valid_dataset
):
    pl = Pipeline.from_config(pipeline_dict)
    trainer = Trainer(pl, train_dataset=train_valid_dataset[0])
    assert trainer.trainer.default_root_dir == str(
        change_to_tmp_working_dir / "training_logs"
    )


def test_deep_copy_of_trainer_config(pipeline_dict, train_valid_dataset):
    pl = Pipeline.from_config(pipeline_dict)
    trainer_config = TrainerConfiguration()
    trainer = Trainer(
        pl, train_dataset=train_valid_dataset[0], trainer_config=trainer_config
    )
    assert trainer_config is not trainer._trainer_config


@pytest.mark.parametrize(
    "input_kwargs,expected_loggers",
    [
        ({}, ["csv", "tensorboard", "wandb"]),
        ({"logger": False}, []),
        (
            {
                "logger": MLFlowLogger(
                    tracking_uri=os.path.join(tempfile.gettempdir(), "mlruns")
                ),
                "add_wandb_logger": False,
            },
            ["csv", "tensorboard", "mlflow"],
        ),
        (
            {
                "logger": [
                    MLFlowLogger(
                        tracking_uri=os.path.join(tempfile.gettempdir(), "mlruns")
                    ),
                    CSVLogger(save_dir=tempfile.gettempdir()),
                ],
                "add_wandb_logger": False,
                "add_tensorboard_logger": False,
            },
            ["csv", "mlflow"],
        ),
    ],
)
def test_add_default_loggers(
    input_kwargs, expected_loggers, pipeline_dict, train_valid_dataset, tmp_path
):
    trainer_config = TrainerConfiguration(
        **input_kwargs, default_root_dir=str(tmp_path)
    )
    trainer = Trainer(
        Pipeline.from_config(pipeline_dict),
        train_dataset=train_valid_dataset[0],
        trainer_config=trainer_config,
    )
    if input_kwargs.get("logger") is not False:
        assert isinstance(trainer.trainer.logger, LoggerCollection)
        assert len(trainer.trainer.logger.experiment) == len(expected_loggers)
    else:
        assert trainer._trainer_config.logger is False

    def loggers_include(logger_type) -> bool:
        return any(
            [
                isinstance(logger, logger_type)
                for logger in trainer._trainer_config.logger
            ]
        )

    for logger in expected_loggers:
        if logger == "csv":
            assert loggers_include(CSVLogger)
        if logger == "tensorboard":
            assert loggers_include(TensorBoardLogger)
        if logger == "wandb":
            assert loggers_include(WandbLogger)
            assert (tmp_path / "wandb").is_dir()
        if logger == "mlflow":
            assert loggers_include(MLFlowLogger)



import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import TrainerConfiguration
from biome.text import VocabularyConfiguration
from biome.text._helpers import PipelineTrainer
from biome.text.errors import EmptyVocabError
from biome.text.features import CharFeatures
from biome.text.features import TransformersFeatures
from biome.text.features import WordFeatures


@pytest.fixture
def pipeline():
    config = {
        "name": "vocab_test",
        "features": {
            "transformers": {"model_name": "sshleifer/tiny-distilbert-base-cased"},
            "word": {"embedding_dim": 2},
            "char": {
                "embedding_dim": 2,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": False,
                },
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": ["good", "bad"],
        },
    }

    return Pipeline.from_config(config)


@pytest.fixture
def train_dataset():
    data = {"text": ["this is a test", "and another one"], "label": ["good", "bad"]}
    return Dataset.from_dict(data)


@pytest.fixture
def valid_dataset():
    data = {
        "text": ["and what about the validation", "do not forget this one"],
        "label": ["bad", "good"],
    }
    return Dataset.from_dict(data)


def test_default_vocab(
    pipeline, train_dataset, valid_dataset, tmp_path, deactivate_pipeline_trainer
):
    # Transformer vocab is added on pipeline creation
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996
    # While word and char vocab should be empty (except for the oov and padding token)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 2
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 2

    # Training should build a default vocab with only the training dataset
    pipeline.train(
        str(tmp_path / "vocab_test_output"),
        training=train_dataset,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996

    # Pretrained pipelines should extend the vocab by default
    pipeline.train(
        str(tmp_path / "vocab_test_output"),
        training=valid_dataset,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 16
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 19
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996


def test_specific_vocab_config(
    pipeline, train_dataset, valid_dataset, deactivate_pipeline_trainer, tmp_path
):
    vocab_config = VocabularyConfiguration(include_valid_data=True)

    pipeline.train(
        output=str(tmp_path / "vocab_test_output"),
        training=train_dataset,
        validation=valid_dataset,
        vocab_config=vocab_config,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 16
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 19
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996


def test_not_touching_vocab(
    pipeline, train_dataset, valid_dataset, tmp_path, deactivate_pipeline_trainer
):
    # vocab_config=None leaves the pipeline's vocab empty from an unpretrained pipeline
    with pytest.raises(EmptyVocabError):
        pipeline.train(
            output=str(tmp_path / "vocab_test_output"),
            training=train_dataset,
            vocab_config=None,
        )

    # vocab_config=None should not extend the vocab for a pretrained pipeline
    pipeline.train(
        output=str(tmp_path / "vocab_test_output"),
        training=train_dataset,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12
    pipeline.train(
        output=str(tmp_path / "vocab_test_output"),
        training=valid_dataset,
        vocab_config=None,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12


def test_restore_vocab(
    pipeline, train_dataset, valid_dataset, tmp_path, deactivate_pipeline_trainer
):
    output = tmp_path / "test_restore_vocab_output"
    pipeline.train(
        output=str(output),
        training=train_dataset,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12

    pipeline.vocab.save_to_files(str(output / "vocabulary"))

    pipeline.train(output=str(output), training=valid_dataset, restore=True)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12



import json

import pytest

from biome.text import Dataset
from biome.text import Pipeline


@pytest.fixture
def dataset() -> Dataset:
    data = {
        "text": ["test", "this", "shaight", "good"],
        "label": ["good", "good", "bad", "good"],
    }
    return Dataset.from_dict(data)


@pytest.fixture
def pipeline(dataset):
    labels = dataset.unique("label")
    return Pipeline.from_config(
        {
            "name": "test_pipeline_evaluate",
            "head": {
                "type": "TextClassification",
                "labels": labels,
            },
        }
    )


def test_pipeline_evaluate(pipeline, dataset, tmp_path):
    prediction_output_file = tmp_path / "prediction_output_file.json"
    metrics = pipeline.evaluate(
        dataset,
        predictions_output_file=str(prediction_output_file),
        batch_size=1,
    )

    assert "loss" in metrics

    predictions = []
    with prediction_output_file.open() as file:
        for line in file.readlines():
            predictions.append(json.loads(line))

    assert len(predictions) == 4
    assert all(["loss" in prediction for prediction in predictions])

    dataset.remove_columns_("label")
    with pytest.raises(ValueError):
        pipeline.evaluate(dataset)



import copy
import json
import logging
import math
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import cast

from allennlp.common import Params
from allennlp.common.checks import ConfigurationError
from allennlp.common.util import prepare_environment
from allennlp.common.util import sanitize
from allennlp.data import PyTorchDataLoader
from allennlp.data.samplers import BucketBatchSampler
from allennlp.models import archive_model
from allennlp.models.archival import CONFIG_NAME
from allennlp.training import GradientDescentTrainer
from allennlp.training import Trainer
from allennlp.training.util import evaluate
from torch.utils.data import IterableDataset

from biome.text import AllenNLPTrainerConfiguration
from biome.text import Pipeline
from biome.text import helpers
from biome.text.dataset import InstanceDataset
from biome.text.model import PipelineModel
from biome.text.training_results import TrainingResults

_LOGGER = logging.getLogger(__name__)


class PipelineTrainer:
    """
    Default trainer for `PipelineModel`

    Parameters
    ----------
    pipeline
        The trainable pipeline
    trainer_config
        The trainer configuration
    output_dir
        The used training folder
    training
        The training instances dataset
    validation
        The validation instances dataset. Optional
    test
        The test instances dataset. Optional
    batch_weight_key
        If non-empty, name of metric used to weight the loss on a per-batch basis.
    epoch_callbacks
        A list of callbacks that will be called at the end of every epoch, and at the start of
        training (with epoch = -1).
    """

    __LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        pipeline: Pipeline,
        trainer_config: AllenNLPTrainerConfiguration,
        output_dir: str,
        training: InstanceDataset,
        validation: Optional[InstanceDataset] = None,
        test: Optional[InstanceDataset] = None,
        batch_weight_key: str = "",
        epoch_callbacks: List["allennlp.training.EpochCallback"] = None,
    ):
        self._pipeline = pipeline
        self._trainer_config = copy.deepcopy(trainer_config)
        self._output_dir = output_dir
        self._batch_weight_key = batch_weight_key
        self._training = training
        self._validation = validation
        self._test = test
        self._epoch_callbacks = epoch_callbacks

        self._setup()

    def _setup(self):
        """Setup the trainer components and local resources"""
        prepare_environment(
            Params(
                {}
                if self._trainer_config.random_seed is None
                else {
                    "random_seed": self._trainer_config.random_seed,
                    "numpy_seed": self._trainer_config.random_seed,
                    "pytorch_seed": self._trainer_config.random_seed,
                }
            )
        )
        os.makedirs(self._output_dir, exist_ok=True)

        serialization_params = sanitize(self._allennlp_configuration())
        with open(os.path.join(self._output_dir, CONFIG_NAME), "w") as param_file:
            json.dump(serialization_params, param_file, indent=4)

        self._pipeline.vocab.save_to_files(os.path.join(self._output_dir, "vocabulary"))

        for dataset in [self._training, self._validation, self._test]:
            if dataset is not None:
                dataset.index_with(self._pipeline.backbone.vocab)

        if self._trainer_config.learning_rate_scheduler is None and (
            self._trainer_config.warmup_steps != 0
            or self._trainer_config.linear_decay is True
        ):
            self._setup_linear_decay_with_warmup()

        trainer_params = Params(
            helpers.sanitize_for_params(self._trainer_config.to_allennlp_trainer())
        )

        pipeline_model = self._pipeline._model

        training_data_loader = create_dataloader(
            self._training,
            self._trainer_config.batch_size,
            self._trainer_config.data_bucketing,
            self._trainer_config.batches_per_epoch,
        )

        validation_data_loader = (
            create_dataloader(
                self._validation,
                self._trainer_config.batch_size,
                self._trainer_config.data_bucketing,
            )
            if self._validation
            else None
        )

        self._trainer = Trainer.from_params(
            model=pipeline_model,
            serialization_dir=self._output_dir,
            data_loader=training_data_loader,
            validation_data_loader=validation_data_loader,
            params=trainer_params,
            epoch_callbacks=self._epoch_callbacks,
        )

    def _setup_linear_decay_with_warmup(self):
        """Setting up the linear_decay_with_warmup learning rate scheduler"""
        if self._trainer_config.linear_decay:
            try:
                num_steps_per_epoch = math.ceil(
                    len(self._training) / self._trainer_config.batch_size
                )
            except TypeError as error:
                if self._trainer_config.training_size is None:
                    raise ConfigurationError(
                        "If you want the learning rate to linearly decay with a lazily loaded training dataset, "
                        "you need to provide the size of the dataset!"
                    ) from error
                else:
                    num_steps_per_epoch = math.ceil(
                        self._trainer_config.training_size
                        / self._trainer_config.batch_size
                    )
        else:
            num_steps_per_epoch = int(1e15)

        self._trainer_config.learning_rate_scheduler = {
            "type": "linear_with_warmup",
            "num_epochs": self._trainer_config.num_epochs,
            "num_steps_per_epoch": num_steps_per_epoch,
            "warmup_steps": self._trainer_config.warmup_steps,
        }

    def test_evaluation(self) -> Dict[str, Any]:
        """
        Evaluates the model against the test dataset (if defined)

        Returns
        -------
        Test metrics information

        """
        test_data = self._test
        if not test_data:
            return {}

        self.__LOGGER.info("The model will be evaluated using the best epoch weights.")
        return evaluate(
            self._pipeline._model,
            data_loader=PyTorchDataLoader(
                test_data, batch_size=self._trainer_config.batch_size
            ),
            cuda_device=self._trainer.cuda_device,
            batch_weight_key=self._batch_weight_key,
        )

    def train(self) -> TrainingResults:
        """
        Train the inner model with given configuration on initialization

        Returns
        -------
        A tuple with trained model path and related metrics information
        """

        from allennlp.models.model import _DEFAULT_WEIGHTS

        for logger in self._epoch_callbacks:
            try:
                logger.init_train(
                    pipeline=self._pipeline,
                    trainer_configuration=self._trainer_config,
                    training=self._training,
                    validation=self._validation,
                    test=self._test,
                )
            except Exception as e:
                self.__LOGGER.warning("Logger %s failed on init_train: %s", logger, e)

        try:
            metrics = self._trainer.train()
        except KeyboardInterrupt:
            # if we have completed an epoch, try to create a model archive.
            if os.path.exists(os.path.join(self._output_dir, _DEFAULT_WEIGHTS)):
                logging.info(
                    "Fine-tuning interrupted by the user. Attempting to create "
                    "a model archive using the current best epoch weights."
                )
                self.save_best_model()
            raise

        for k, v in self.test_evaluation().items():
            metrics["test_" + k] = v

        self.save_best_model()

        with open(os.path.join(self._output_dir, "metrics.json"), "w") as metrics_file:
            metrics_json = json.dumps(metrics, indent=2)
            metrics_file.write(metrics_json)

        training_results = TrainingResults(
            os.path.join(self._output_dir, "model.tar.gz"), metrics
        )

        for logger in self._epoch_callbacks:
            try:
                logger.end_train(training_results)
            except Exception as e:
                self.__LOGGER.warning("Logger %s failed on end_traing: %s", logger, e)

        return training_results

    def save_best_model(self):
        """Packages the best model as tar.gz archive"""
        archive_model(self._output_dir)

    def _allennlp_configuration(self) -> Dict[str, Any]:
        """Creates an allennlp configuration for pipeline train experiment configuration"""
        allennlp_config = {
            # TODO: this does not make much sense, we should save our trainer configuration format!
            "trainer": self._trainer_config.to_allennlp_trainer(),
            "model": {
                "config": self._pipeline.config.as_dict(),
                "type": PipelineModel.__name__,
            },
        }

        return copy.deepcopy({k: v for k, v in allennlp_config.items() if v})


def create_dataloader(
    dataset: InstanceDataset,
    batch_size: int,
    data_bucketing: bool = False,
    batches_per_epoch: Optional[int] = None,
) -> PyTorchDataLoader:
    """Returns a pytorch DataLoader for AllenNLP

    Parameters
    ----------
    dataset
        The data set for the DataLoader
    batch_size
        Size of the batch.
    data_bucketing
        If enabled, try to apply data bucketing over training batches.
    batches_per_epoch
        Determines the number of batches after which an epoch ends.
        If the number is smaller than the total amount of batches in your data,
        the second "epoch" will take off where the first "epoch" ended.
        If this is `None`, then an epoch is set to be one full pass through your data.

    Returns
    -------
    data_loader
    """
    return (
        PyTorchDataLoader(
            dataset,
            batch_sampler=BucketBatchSampler(
                data_source=dataset, batch_size=batch_size
            ),
            batches_per_epoch=batches_per_epoch,
        )
        if data_bucketing and not isinstance(dataset, IterableDataset)
        else PyTorchDataLoader(
            dataset, batch_size=batch_size, batches_per_epoch=batches_per_epoch
        )
    )


def create_trainer_for_finding_lr(
    model: PipelineModel,
    trainer_config: AllenNLPTrainerConfiguration,
    training_data: InstanceDataset,
) -> GradientDescentTrainer:
    """Returns an AllenNLP Trainer used for the learning rate scan.

    Parameters
    ----------
    model
        The underlying model
    trainer_config
        A trainer configuration
    training_data
        The training data
    """
    prepare_environment(Params({}))

    trainer_params = Params(
        helpers.sanitize_for_params(trainer_config.to_allennlp_trainer())
    )

    training_data_loader = create_dataloader(
        training_data, trainer_config.batch_size, trainer_config.data_bucketing
    )

    return cast(
        "GradientDescentTrainer",
        Trainer.from_params(
            model=model,
            data_loader=training_data_loader,
            params=trainer_params,
            serialization_dir=None,
        ),
    )



import json
import random
from typing import Tuple

import numpy as np
import pytest
import torch

from biome.text import AllenNLPTrainerConfiguration
from biome.text import Dataset
from biome.text import Pipeline
from biome.text import VocabularyConfiguration
from biome.text.configuration import CharFeatures
from biome.text.configuration import WordFeatures


@pytest.fixture
def train_valid_dataset(resources_data_path) -> Tuple[Dataset, Dataset]:
    """Returns both training and validation datasets"""

    training_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.train.csv")
    )
    validation_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.valid.csv")
    )

    return training_ds, validation_ds


@pytest.fixture
def pipeline_dict() -> dict:
    """Pipeline config dict. You need to update the labels!"""

    pipeline_dictionary = {
        "name": "german_business_names",
        "features": {
            "word": {"embedding_dim": 16, "lowercase_tokens": True},
            "char": {
                "embedding_dim": 16,
                "encoder": {
                    "type": "gru",
                    "num_layers": 1,
                    "hidden_size": 32,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
            "pooler": {
                "type": "gru",
                "num_layers": 1,
                "hidden_size": 16,
                "bidirectional": True,
            },
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [16],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }

    return pipeline_dictionary


@pytest.fixture
def trainer_dict() -> dict:
    """Returns the trainer dictionary"""

    return {
        "batch_size": 64,
        "num_epochs": 5,
        "optimizer": {"type": "adam", "lr": 0.01},
        "cuda_device": -1,
    }


def test_text_classification(
    tmp_path, pipeline_dict, trainer_dict, train_valid_dataset
):
    """Apart from a well specified training, this also tests the vocab creation!"""

    random.seed(42)
    np.random.seed(422)
    torch.manual_seed(4222)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(4222)

    pl = Pipeline.from_config(pipeline_dict)
    train_ds = train_valid_dataset[0]
    valid_ds = train_valid_dataset[1]
    trainer = AllenNLPTrainerConfiguration(**trainer_dict)
    vocab_config = VocabularyConfiguration(max_vocab_size={"word": 50})

    output = tmp_path / "output"

    pl.train(
        output=str(output),
        trainer=trainer,
        training=train_ds,
        validation=valid_ds,
        vocab_config=vocab_config,
    )

    evaluation = pl.evaluate(valid_ds)
    assert evaluation["loss"] == pytest.approx(0.873307535648346, abs=0.003)

    assert pl.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl.vocab.get_vocab_size(CharFeatures.namespace) == 83

    assert pl.num_trainable_parameters == 22070

    with (output / "metrics.json").open() as file:
        metrics = json.load(file)

    # It may fail in some systems
    assert metrics["training_loss"] == pytest.approx(0.684, abs=0.003)

    # Test vocab from a pretrained file
    pl = Pipeline.from_pretrained(str(output / "model.tar.gz"))

    assert pl.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl.vocab.get_vocab_size(CharFeatures.namespace) == 83



import pytest
import torch

from biome.text import AllenNLPTrainerConfiguration
from biome.text import Dataset
from biome.text import Pipeline


@pytest.fixture(scope="module")
def dataset() -> Dataset:
    data = {"text": ["Test", "this", "shaight", "!"], "label": ["0", "1", "0", "0"]}

    return Dataset.from_dict(data)


@pytest.fixture(scope="module")
def pipeline(dataset) -> Pipeline:
    config = {
        "name": "test_trainer",
        "features": {"word": {"embedding_dim": 2}},
        "head": {"type": "TextClassification", "labels": list(set(dataset["label"]))},
    }
    pl = Pipeline.from_config(config)

    return pl


@pytest.mark.skipif(
    torch.cuda.device_count() < 1, reason="Using AMP requires a cuda device"
)
def test_use_amp(dataset, pipeline, tmp_path, capsys):
    trainer_config = AllenNLPTrainerConfiguration(
        num_epochs=1,
        batch_size=2,
        use_amp=True,
    )

    pipeline.train(
        output=str(tmp_path / "test_use_amp_output"),
        training=dataset,
        trainer=trainer_config,
    )

    captured = capsys.readouterr()
    assert "use_amp = True" in captured.err



import dataclasses
import logging
import os
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from allennlp.training import EpochCallback
from allennlp.training import GradientDescentTrainer
from mlflow.entities import Experiment
from mlflow.tracking import MlflowClient
from mlflow.utils import mlflow_tags

from biome.text.dataset import InstanceDataset
from biome.text.training_results import TrainingResults

# We do not require wandb
_HAS_WANDB = False
try:
    import wandb
    from pkg_resources import parse_version

    assert parse_version(wandb.__version__) >= parse_version("0.10.12")
except ImportError:
    pass
except AssertionError:
    logging.warning("To log to WandB, please update your wandb client to >=0.10.12")
else:
    wandb.ensure_configured()
    _HAS_WANDB = True


class BaseTrainLogger(EpochCallback):
    """Base train logger for pipeline training"""

    def init_train(
        self,
        pipeline: "Pipeline",
        trainer_configuration: "TrainerConfiguration",
        training: InstanceDataset,
        validation: Optional[InstanceDataset] = None,
        test: Optional[InstanceDataset] = None,
    ):
        """Init train logging

        Parameters
        ----------
        pipeline
            The training pipeline
        trainer_configuration
            The trainer configuration
        training
            Training dataset
        validation
            Validation dataset
        test
            Test dataset
        """
        pass

    def end_train(self, results: TrainingResults):
        """End train logging

        Parameters
        ----------
        results
            The training result set
        """
        pass

    def log_epoch_metrics(self, epoch: int, metrics: Dict[str, Any]):
        """Log epoch metrics

        Parameters
        ----------
        epoch
            The current epoch
        metrics
            The metrics related to current epoch
        """
        pass

    def __call__(
        self,
        trainer: GradientDescentTrainer,
        metrics: Dict[str, Any],
        epoch: int,
        is_master: bool,
    ):
        if epoch >= 0:
            self.log_epoch_metrics(epoch, metrics)


class MlflowLogger(BaseTrainLogger):
    """A common mlflow logger for pipeline training

    Parameters
    ----------
    experiment_name
        The experiment name
    artifact_location
        The artifact location used for this experiment
    run_name
        If specified, set a name to created run
    tags
        Extra arguments used as tags to created experiment run
    """

    __LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        experiment_name: str = None,
        artifact_location: str = None,
        run_name: str = None,
        **tags,
    ):
        self._client = MlflowClient()
        self._experiment = self._configure_experiment_with_retry(
            experiment_name, artifact_location
        )

        tags = tags or {}
        if run_name:
            tags[mlflow_tags.MLFLOW_RUN_NAME] = run_name

        run = self._client.create_run(self._experiment.experiment_id, tags=tags)
        self._run_id = run.info.run_id

        self._skipped_metrics = ["training_duration"]

    def _configure_experiment_with_retry(
        self, experiment_name: str, artifact_location: str, retries: int = 5
    ) -> Optional[Experiment]:
        """Tries to configure (fetch or create) an mlflow experiment with retrying process on errors"""
        if retries <= 0:
            return None
        try:
            experiment = self._client.get_experiment_by_name(
                experiment_name or "default"
            )
            if experiment:
                return experiment

            return self._client.get_experiment(
                self._client.create_experiment(experiment_name, artifact_location)
            )
        except Exception as e:
            self.__LOGGER.debug(e)
            return self._configure_experiment_with_retry(
                experiment_name, artifact_location, retries=retries - 1
            )

    def init_train(
        self,
        pipeline: "Pipeline",
        trainer_configuration: "TrainerConfiguration",
        training: InstanceDataset,
        validation: Optional[InstanceDataset] = None,
        test: Optional[InstanceDataset] = None,
    ):
        from pandas import json_normalize

        for prefix, params_set in [
            ("pipeline", json_normalize(pipeline.config.as_dict())),
            ("trainer", json_normalize(dataclasses.asdict(trainer_configuration))),
        ]:
            for key, value in params_set.to_dict(orient="records")[0].items():
                if value:
                    self._client.log_param(self._run_id, f"{prefix}.{key}", value)
        self._client.log_param(
            self._run_id, key="pipeline.num_parameters", value=pipeline.num_parameters
        )
        self._client.log_param(
            self._run_id,
            key="pipeline.num_trainable_parameters",
            value=pipeline.num_trainable_parameters,
        )

    def log_epoch_metrics(self, epoch: int, metrics: Dict[str, Any]):

        [
            self._client.log_metric(self._run_id, key=k, value=v, step=epoch)
            for k, v in metrics.items()
            if k not in self._skipped_metrics
        ]

    def end_train(self, results: TrainingResults):
        try:
            self._client.log_artifact(self._run_id, local_path=results.model_path)
            [
                self._client.log_metric(self._run_id, key=k, value=v)
                for k, v in results.metrics.items()
                if k not in self._skipped_metrics
            ]
        finally:
            self._client.set_terminated(self._run_id)


class WandBLogger(BaseTrainLogger):
    """Logger for WandB

    Parameters
    ----------
    project_name
        Name of your WandB project
        The environment variable 'WANDB_PROJECT' will override this value.
    run_name
        Name of your run
    tags
        Extra arguments used as tags to created experiment run
    """

    def __init__(
        self, project_name: str = "biome", run_name: str = None, tags: List[str] = None
    ):
        if wandb.api.api_key is None:
            wandb.termwarn(
                "W&B installed but not logged in. "
                "Run `wandb login` or `import wandb; wandb.login()` or set the WANDB_API_KEY env variable."
            )

        self.project_name = os.environ.get("WANDB_PROJECT") or project_name
        self.run_name = run_name
        self.tags = tags

        self._run = None

    def init_train(
        self,
        pipeline: "Pipeline",
        trainer_configuration: "TrainerConfiguration",
        training: InstanceDataset,
        validation: Optional[InstanceDataset] = None,
        test: Optional[InstanceDataset] = None,
    ):
        config = {
            "pipeline": pipeline.config.as_dict(),
            "trainer": dataclasses.asdict(trainer_configuration),
        }
        config["pipeline"]["num_parameters"] = pipeline.num_parameters
        config["pipeline"][
            "num_trainable_parameters"
        ] = pipeline.num_trainable_parameters

        Path(".wandb").mkdir(exist_ok=True)
        self._run = wandb.init(
            project=self.project_name,
            name=self.run_name,
            tags=self.tags,
            config=config,
            dir=".wandb",
        )

    def log_epoch_metrics(self, epoch: int, metrics: Dict[str, Any]):
        wandb.log(metrics)

    def end_train(self, results: TrainingResults):
        if self._run:
            self._run.finish()


def is_wandb_installed_and_logged_in() -> bool:
    """Checks if wandb is installed and if a login is detected.

    Returns
    -------
    bool
        Is true, if wandb is installed and a login is detected, otherwise false.
    """
    if not _HAS_WANDB:
        return False
    if wandb.api.api_key is None:
        wandb.termwarn(
            "W&B installed but not logged in. "
            "Run `wandb login` or `import wandb; wandb.login()` or set the WANDB_API_KEY env variable."
        )
        return False
    return True


def add_default_wandb_logger_if_needed(
    loggers: List[BaseTrainLogger],
) -> List[BaseTrainLogger]:
    """Adds the default WandBLogger if a WandB login is detected and no WandBLogger is found in `loggers`.

    Parameters
    ----------
    loggers
        List of loggers used in the training

    Returns
    -------
    loggers
        List of loggers with a default WandBLogger at position 0 if needed
    """
    if any([isinstance(logger, WandBLogger) for logger in loggers]):
        pass
    elif not is_wandb_installed_and_logged_in():
        pass
    else:
        loggers = [WandBLogger()] + loggers

    return loggers
